#!/usr/bin/perl
use strict;
use warnings;
use SlidingWindowSearch qw(ListMatches CountMatches ParsePatternSpecFile);
use Getopt::Long;
use Pod::Usage;
use Time::HiRes;

my $opt_debug = 0;
my $opt_silent = 0;
my $opt_checks = 1;
my $opt_time = 0;
my $startTime;

### Read command-line options

sub ProcessCommandLine() {
	my $seqFile;
	my $seqFromStdin;
	my $patFile;
	my $method;
	my $opt_help;
	my $opt_man;

	$ARGV[-1] =~ s/\r//;     # in case called from script written on Windows
	Getopt::Long::Configure("bundling");
	GetOptions(
		'patfile|p=s' 	=> \$patFile,
		'seqfile|s|i=s' => \$seqFile,
		'' 				=> \$seqFromStdin,
		'use|u=s'		=> \$method,
		'debug|D!' 		=> \$opt_debug,
		'silent|S!'		=> \$opt_silent,
		'checks|C!'		=> \$opt_checks,
		'time|T!'		=> \$opt_time,
		'help|h'		=> \$opt_help,
		'man|m'			=> \$opt_man,		
	) or pod2usage(-verbose => 1) && exit;	# based on online examples
	if ($opt_man) {pod2usage(-verbose => 2) && exit};
	if ($opt_help) {pod2usage(-verbose => 1) && exit};
	
	unless (defined($patFile)) {
			print STDERR "No pattern file specified\n";
			$opt_help = 1;
	}
	unless (defined($seqFile) or $seqFromStdin) {
			print STDERR "No sequence file specified\n";
			$opt_help = 1;
	}
	if ($opt_help) {print STDERR "\n"; pod2usage(-verbose => 1) && exit};
	
	if (defined($seqFile) and '-' eq $seqFile) {$seqFromStdin = 1};
	unless ($seqFromStdin or -e $seqFile) {
			die("\n*** Execution of $0 terminating:\n  Sequence file $seqFile could not be found\n");
	}
	unless (-e $patFile) {
			die("\n*** Execution of $0 terminating:\n  Pattern file $patFile could not be found\n");
	}

	SlidingWindowSearch::SetDebug($opt_debug);
	SlidingWindowSearch::SetSilent($opt_silent);
	SlidingWindowSearch::SetChecks($opt_checks);
	return($patFile,$seqFile,$seqFromStdin,$method);
}
		
### Read sequence
#     Very limited routine to read sequence:
#     All lines starting with A-Z are presumed to contain sequence
#     Internal whitespace is removed
#     Presence of characters other than A-Z or whitespace will throw a fatal error

sub GetSequence($$) {
	(my $seqFile, my $seqFromStdin) = @_;
	my $seq_fh;
	if ($seqFromStdin) {
		$seq_fh = *STDIN;
		unless ($opt_silent) {print "\nSequence from STDIN\n"};
	} else {
		open($seq_fh, "<", $seqFile)
			or die "cannot open < $seqFile: $!";
		unless ($opt_silent) {print "\nSequence file '$seqFile'\n"};
	}

	my $seq='';
	my $lineCount = 0;
	while (defined(my $line = <$seq_fh>)) {
		$lineCount++;
		next unless $line =~ /^[A-Z]/;
		chomp($line);
		unless ($line =~ /^([A-Z\s]+)$/) {
			while ($line =~ /[^A-Z\s]/g) {;
				my $pos = pos($line);
				die("\n*** Execution of $0 terminating:\n  Sequence contained unexpected character at line $lineCount position $pos\n");
			}
		}
		my $seqLine = $1;
		$seqLine =~ s/\s+//g;
		$seq .= $seqLine;
	}

	unless ($seqFromStdin) {
		close($seq_fh);
	}
	my $seqLen = length($seq);
	unless ($opt_silent) {print "  Sequence length $seqLen\n"};
	return($seq);
}

### Read and parse MHC pattern file
#     File is read into memory, then passed to module routine for parsing
sub GetPatterns($) {
	(my $patFile, undef) = @_;

	open(my $mhc_fh, "<", $patFile)
		or die "cannot open < $patFile: $!";
	my @lines = <$mhc_fh>;
	close($mhc_fh);

	(my $iFail, my @results) = (ParsePatternSpecFile($patFile,@lines));
	if ($iFail) {
		(my $lineNo, my $failMsg) = @results;
		if ($iFail < 40) {
			die("\n*** Execution of $0 terminating:\n Error in parsing line $lineNo of pattern specification file \"$patFile\"\n -- $failMsg")
		} else {
			die("\n*** Execution of $0 terminating:\n Error in parsing pattern specification file \"$patFile\"\n -- $failMsg")
		}
	}
	
	return(@results);
}


### main script is here
sub main() {
	(my $patFile, my $seqFile, my $seqFromStdin, my $method) = ProcessCommandLine();
	(my $patterns_ref, my $patWidth, my $patWidthExtended) = GetPatterns($patFile);
	my $seq = GetSequence($seqFile, $seqFromStdin);

	### Count matches
	if ($opt_time) {$startTime = Time::HiRes::clock()};
	my %matchCount = CountMatches($seq,@$patterns_ref,$patWidth,$patWidthExtended,$method);
	if ($opt_time) {print "\n", Time::HiRes::clock() - $startTime, " process seconds elapsed\n"};
	print "\nSequence pattern match counts:\n";
	foreach my $pattern (sort {$matchCount{$b} <=> $matchCount{$a} or $a cmp $b} keys %matchCount) {
		print "  $pattern \t$matchCount{$pattern}\n";
	}
}

main();

__END__

=head1 NAME

 MHC_match.pl
 
=head1 SYNOPSIS

 MHC_match.pl -s SEQUENCE -p PATTERNS
 
=head1 DESCRIPTION

 Given one or more MHC binding motifs specified in a user-supplied 
 patterns file, list fixed-length sequence extracts from a 
 user-supplied sequence file that match these motifs, with a count 
 of the number of motifs that each sequence extract matches.
 
=head1 ARGUMENTS

 -s, --seqfile SEQUENCE
 -p, --patfile PATTERNS 
 - , -s -       take sequence data from STDIN
 
 For details of SEQUENCE and PATTERNS file formats, see full man page
 
=head1 OPTIONS

 -D, --debug    print additional run-time information to STDERR, 
                   including the number of matches for each pattern
 -S, --silent   minimal output (only the results)
 -C, --checks   check sequence for non-alpha characters (default)
 -T, --time     report time taken on main loop
 -h, --help     summary of arguments and options
 -m, --man      man page
 -u, --use <method>
 
 Recognised <method>s are 'UsingGrep' and 'SlidingWindow' (default)
 
=head1 FILE FORMATS

=head2 SEQUENCE FILES

 All lines starting with A-Z are presumed to contain sequence
 Internal whitespace is removed
 Presence of characters other than A-Z or whitespace will be
 rejected and the program will stop (unless the --nochecks
 option has been specified).
 
=head2 PATTERN FILES

 Before any patterns are specified, specification lines must appear
 giving the pattern width and (optionally) extended pattern width.
 These specification lines start with a *, eg:

    * Pattern width = 9
    * Extended pattern width = 15

These lines may be repeated, but the values may not be changed.

Pattern specification files may specify a number of patterns, each
separated by a blank line.  The specifications should otherwise be
contiguous: blank lines will be taken as signifying the start of a
new pattern.

Each pattern is specified by giving the test position, followed by
the character values allowed at that position, eg:

    1 = A, G, P
    3 = N, D

Commas and spaces are optional; a colon may be used instead of the '='

Lines beginning with # will be treated as comments, and ignored.
Such lines may contain eg the names of MHC allele binding motifs
from which each pattern has been derived.
 
=head1 EXAMPLES

 perl MHC_match.pl -p mhc_class1.txt -s testSeq.txt
 cat testSeq.txt | perl MHC_match.pl -p mhc_class1.txt -
 perl MHC_match.pl --nochecks -T -p mhc_class1.txt -s testSeq.txt -u UsingGrep
 
=head1 AUTHOR

James Heald E<lt>j.heald@ucl.ac.ukE<gt>

=head1 HISTORY

Version 1.0,  19 November 2012

=head1 COPYRIGHT AND LICENSE

Copyright 2012 by James Heald

This script is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 
 
=cut
 

