#!/usr/bin/perl
package SlidingWindowSearch;
use strict;
use warnings;
use Exporter; 
use Try::Tiny; 
use Carp;

our $VERSION = 0.99;
our @ISA = qw(Exporter); 
our @EXPORT_OK = qw(ListMatches CountMatches ParsePatternSpecFile);
our %EXPORT_TAGS = (
    all => \@EXPORT_OK,
);

					
## Global variable $debug == 1 for verbose output
my $debug = 0;
sub SetDebug($) {
	($debug, undef) = @_;
}

sub GetDebug() {
	return($debug);
}

## Global variable $silent == 1 for minimal output
my $silent = 0;
sub SetSilent($) {
	($silent, undef) = @_;
}

sub GetSilent() {
	return($silent);
}

## Global variable $checks == 1 if sequence should be checked for non [A-Z] characters
my $checks = 1;
sub SetChecks($) {
	($checks, undef) = @_;
}

sub GetChecks() {
	return($checks);
}

#### Simple version of ListMatches, using dice & match method suggested in the problem spec 
## List sub-strings of $seq matching pattern $pattern	
sub ListMatches_UsingGrep($$$;$) {
	(my $seq, my $pattern, my $patWidth, my $patWidthExtended) = @_;
	$patWidthExtended //= $patWidth;     # use $patWidth if $patWidthExtended not given
	return grep {/$pattern/} map {substr($seq,$_,$patWidthExtended)} 0..(length($seq)-$patWidthExtended);
}
	
#### Alternate approach, in effect running a regex over the original $seq with a sliding window	
## Proof of concept: simple case
sub ListMatches_HardCoded_Simple($) {		
	my $seq = $_[0];
	my @matches = ($seq =~ /.(?=(.A.))/g);
	
	# edge case: check for match in position zero
	if ($seq =~ /^(?=(.A.))/g) {
		unshift(@matches,$1);
	}
	@matches;
}

## Proof of concept: extended case
sub ListMatches_HardCoded_Extended($) {
	my $seq = $_[0];
	
	# match with up to $overhang free characters either side
	my @matches = ($seq =~ /.(?=(.{0,2}A.{0,2}))/g);

	# edge case: check for match in position zero
	if ($seq =~ /^(?=(.{0,2}A.{0,2}))/g) { 
		unshift(@matches,$1);
	}

	# can't have $overhang characters on both sides, so limit to max width
	@matches = map(substr($_,0,3),@matches);

	# edge case: last $overhang matches may not be full width
	my $lastIndex = $#matches;
	for (my $i=$lastIndex;$i>$lastIndex-2;$i--) {
		if (length($matches[$i]) < 3) {
			delete $matches[$i];
		} else {
			last;    # can stop looking back, as soon as we've got one full-width
		}
	}
	@matches;
}

sub ListMatches($$;$$) {    
	(my $seq, my $pattern, my $patWidth, my $patWidthExtended) = @_;
	if ($debug) {print STDERR "  Running pattern '$pattern'\n"};
	
	my $patOverhang = 0;
	if (defined($patWidthExtended)) {
		# allow up to $patOverhang free characters either side
		$patOverhang = $patWidthExtended - $patWidth;
		if ($debug) {print STDERR "  Overhang = $patOverhang\n"};
		$pattern = ".{0,$patOverhang}" . $pattern . ".{0,$patOverhang}";
	}
	
	my $regex;
	try {
		$regex = qr($pattern);
	} catch {
		croak "\n*** Execution of $0 terminating:\n  Unable to parse pattern '$pattern'\n    $_";
	};
	my @matches = ($seq =~ /.(?=($regex))/g);

	# edge case: check for match in position zero
	if ($seq =~ /^(?=($regex))/g) {
		unshift(@matches,$1);
	}
	
	if ($patOverhang) {
		# limit to max width (otherwise might have up to $patOverhang characters on both sides) 
		@matches = map(substr($_,0,$patWidthExtended),@matches);

		# edge case: last $patOverhang matches may not be full width
		my $lastIndex = $#matches;
		for (my $i=$lastIndex;$i>=0 and $i>=$lastIndex-$patOverhang;$i--) {
			my $len = length($matches[$i]);
			if (length($matches[$i]) < $patWidthExtended) {
				delete $matches[$i];
			} else {
				last;    # can stop looking back, as soon as we've got one full-width
			}
		}
	}
	
	if ($debug) {
		my $nmatches = scalar(@matches);
		print STDERR "  $nmatches matches found (including duplicates)\n";
	}
	@matches;
}

## uniq() -- taken from List::MoreUtils, and Perl Cookbook
sub uniq (@) {
    my %seen = ();
    grep { not $seen{$_}++ } @_;
}

# Dispatch table of alternative subrs for doing the ListMatches
# Code influenced by 'chromatic' (2012), "Modern Perl", p. 83
my %dispatch =
(
	SlidingWindow	=> \&ListMatches,
	UsingGrep		=> \&ListMatches_UsingGrep,
);

#### Main driver routine
## Find sub-strings of $seq matching the patterns in $patList_Ref, counting how many patterns each sub-string matches
sub CountMatches($\@$;$$\%) {
	(my $seq, my $patList_Ref, my $patWidth, my $patWidthExtended, my $method, my $matchCount_Ref) = @_;
	my %matchCount = %{$matchCount_Ref // {}};
	if ($checks) {SeqCheck($seq)};

	if (defined($method) && (!exists $dispatch{$method})) {
		carp("Method '$method' not recognised; reverting to default");
		undef($method);
	}
	$method //= 'SlidingWindow';   # default

	if ($debug) {print STDERR "\nRunning pattern comparisons ($method):\n"};
	foreach my $pattern (@{$patList_Ref}) {
		my @matches = uniq($dispatch{$method}->($seq,$pattern,$patWidth,$patWidthExtended));
		if ($debug) {
			my $nmatches = scalar(@matches);
			print STDERR "  $nmatches unique matches found\n\n";
		}
		map {$matchCount{$_}++} @matches;
	}
	return (%matchCount);
}

sub SeqCheck($) {
	(my $seq, undef) = @_;
	if (0 == length($seq)) {
		die ("\n*** Execution of $0 terminating:\n  Sequence has zero length.\n");
	}
	while ($seq =~ /[^A-Z]/g) {
		my $pos = pos($seq);
		die ("\n*** Execution of $0 terminating:\n  Unrecognised character found at sequence position $pos. Only characters A-Z expected.\n");
	}
}

#### Parse the lines of a pattern specification file
sub ParsePatternSpecFile($@) {
	(my $filename, my @lines) = @_;
	
	# overloading: allow either a real array or a ref to an array to be passed
	my $lines_ref;
	$lines_ref = ref($lines[0]) ? $lines[0] : \@lines;
	
	push(@$lines_ref,'');  # add extra blank line to force recording of final record
	
	my @patterns;
	my $patWidth;
	my $patWidthExtended;
	
	unless ($silent) {print "\nPattern file '$filename'\n"};
	
	my $currentPattern;
	my $nextPos = 1;
	my $lineCount = 0;
	foreach my $line (@$lines_ref) {
		$lineCount++;
		chomp($line);
		if ($line =~ /^\*\s*PATTERN WIDTH\s*[:=]\s*(\d+)\s*$/i) {
			if (defined($patWidth) and ($patWidth != $1)) {
				return(10,$lineCount,"Pattern Width cannot be re-defined\n");
			}
			$patWidth = $1;
		}
		elsif ($line =~ /^\*\s*EXTENDED PATTERN WIDTH\s*[:=]\s*(\d+)\s*$/i) {
			if (defined($patWidthExtended) and ($patWidthExtended != $1)) {
				return(11,$lineCount,"Extended Pattern Width cannot be re-defined\n");
			}
			$patWidthExtended = $1;
			unless (defined($patWidth)) {
				return(12,$lineCount,"Pattern Width must be specified before Extended Pattern Width\n");
			}
			unless ($patWidthExtended >= $patWidth) {
				return(13,$lineCount,"Extended Pattern Width must not be less than Pattern Width\n");
			}
		}
		elsif ($line =~ /^#/) {
			next;
		}
		elsif ($line =~ /^([0-9]+)\s*[:=]\s*(([A-Z](,\s*)?)+)\s*$/i) {
			my $pos = $1;
			my $matchChars = uc($2);
			$matchChars =~ s/,\s*//g;
			unless ($pos > 0) 
				{return(21,$lineCount,"Test position must be greater than zero\n")}
			unless ($pos >= $nextPos)
				{return(22,$lineCount,"Unexpected test position $pos (new records must be separated by blank lines)\n")};
			unless (defined($patWidth)) 
				{return(23,$lineCount,"Pattern Width must be specified before Patterns can be read\n")};
			unless ($pos <= $patWidth)
				{return(24,$lineCount,"Test position must not exceed pattern width\n")}
			
			$currentPattern .= '.' x ($pos - $nextPos);
			$currentPattern .= '[' . $matchChars . ']';
			$nextPos = $pos + 1;
		}
		elsif ($line =~ /^\s*$/) {
			next unless ($currentPattern);
			unless ($nextPos > $patWidth) {
				$currentPattern .= '.' x ($patWidth + 1 - $nextPos);
			}
			push(@patterns, $currentPattern);
			unless ($silent) {
				print "  Pattern: '$currentPattern'  \tWidth: $patWidth";
				if ($patWidthExtended) {print "   Extended Width: $patWidthExtended"};
				print "\n";
			}
			$currentPattern = '';
			$nextPos = 1;
		} else {
			return(30,$lineCount,"Could not interpret line '$line'\n");
		}
	}
	unless (scalar @patterns) {
		return(40,$lineCount,"No patterns found in file\n");
	}
	return (0, \@patterns, $patWidth, $patWidthExtended);
}	
		
1;

__END__

=head1 NAME

SlidingWindowSearch - search a string for specified patterns using a sliding window 

=head1 SYNOPSIS

  use SlidingWindowSearch qw(ListMatches CountMatches ParsePatternSpecFile)
  @matchList = ListMatches($seq,$pattern)
  @matchList = ListMatches($seq,$pattern,$patWidth,$patWidthExtended)
  %matchCount = CountMatches($seq,@patList)
  %matchCount = CountMatches($seq,@patList,
       $patWidth,$patWidthExtended,$method,%matchCount)
  ($iFail, $patterns_ref,$patWidth,$patWidthExtended) 
       = ParsePatternSpecFile($filename,@lines)

=head1 DESCRIPTION

B<SlidingWindowSearch> provides routines to search a string for substrings matching
a specified pattern or patterns, using a sliding window approach.  The matching
substrings may be specified to either match the length of the supplied pattern(s)
exactly; or alternatively, if an "extended pattern length" is specified, then any
substring of this length which contains the specified pattern will be considered to
match.
 
A routine is also provided to parse text files containing specifications of the
patterns.
 
One application of the routines is to identify polypeptide fragments created by
the action of protease enzymes on proteins that can be recognised and displayed
by specified MHC Class I and Class II receptors.

The following main routines are provided:

=over 4

=item @matchList = ListMatches($seq,$pattern,$patWidth,$patWidthExtended)

Returns a list of the sub-strings of C<$seq> that match the pattern $pattern
(specified as a Perl regular expression).  

C<$patWidth> specifies the width of
the pattern, i.e. the number of characters of the substring to be matched.
C<$patWidthExtended> specifies the width of the sub-string sought such that
a match of the pattern at any position within the sub-string should be accepted.

For the standard routine provided, C<$patWidth> and C<$patWidthExtended> are required
only if the extended pattern matching described above is sought.

An alternative implementation, C<SlidingWindowSearch::ListMatches_UsingGrep>, 
is also available.  For that implementation, one or both of either 
C<$patWidth> or C<$patWidthExtended> must always be specified.

Matching sub-strings are returned in the order in which they occur in the full
string, with duplicates included.

=item %matchCount = CountMatches($seq,@patList,$patWidth,$patWidthExtended)

Returns a hash containing the sub-strings that match one or more of the 
supplied list of patterns C<@patList>, and how many of those patterns
each sub-string matched.  (Multiple matches to the same pattern are 
suppressed).

The arguments C<$patwidth> and C<$patWidthExtended> have the same meanings
as for C<ListMatches> above.

Additional arguments C<$method> and C<%matchCount> can also be supplied.
C<$method> is a text string specifying the implementation of C<ListMatches>
to be used.  Currently acceptable arguments are 'SlidingWindow' (the default),
and 'UsingGrep' (as described above).

The optional final argument C<%matchCount> allows an existing hash
C<%matchCount> from a previous call to the subroutine to be updated.

=item ParsePatternSpecFile($filename,@lines)

Parses lines that have been read from a pattern specification file,
returning a list C<($iFail,$patterns_ref,$patWidth,$patWidthExtended)>
containing an exit state; a reference to a list of the patterns 
themselves, translated into Perl regular expressions; the specified
width of the patterns, and any specified extended width for the 
patterns.  (If no such extended width has been specified, this 
entry will be undefined).

An exit state of 0 indicates success.  Non-zero exit indicates
failure, which is reported with the line number of the specification
file, and an error message. (See code for details).

The variable C<$filename> is used only for convenience, to report
that processing of the file is underway.

=back

=head2 ADDITIONAL CONVENIENCE ROUTINES

=over 4

=item SlidingWindowSearch::SetDebug($debugLevel)

=item SlidingWindowSearch::GetDebug()

Set and Get internal global variable C<$debug>.

C<SetDebug(1)> produces more detailed progress information, including
a count of the numbers of sub-strings matching each pattern.

=item SlidingWindowSearch::SetSilent($silentLevel)

=item SlidingWindowSearch::GetSilent()

Set and Get internal global variable C<$silent>.

C<SetSilent(1)> prevents the logging of certain information, including
a listing of the patterns to be matched, and the length of the string
to be analysed.

=item SlidingWindowSearch::SetChecks($checksLevel)

=item SlidingWindowSearch::GetChecks()

Set and Get internal global variable C<$checks>.

C<SetChecks(0)> prevents checking for blanks, control codes
and other non-[A-Z] characters in the supplied string to be
searched.

=back

=head2 FORMAT FOR PATTERN SPECIFICATION FILES

Pattern specification files may specify a number of patterns, each
separated by a blank line.  The specifications should otherwise be
contiguous: blank lines will be taken as signifying the start of a
new pattern.

Lines beginning with # will be treated as comments, and ignored.
Such lines may contain eg the names of MHC allele binding motifs
from which each pattern has been derived.

Specification lines specifying the pattern width and (optionally)
the extended pattern width must appear before any patterns are 
specified.  These specification lines start with a *, eg:

    * Pattern width = 9
    * Extended pattern width = 15

These lines may be repeated, but the values may not be changed.

Each pattern is specified by giving the test position, followed by
the character values allowed at that position, eg:

    1 = A, G, P
    3 = N, D

Commas and spaces are optional; a colon may be used instead of the '='

=head1 EXPORTS

No symbols are exported by default.  

Routines C<ListMatches>, C<CountMatches> and C<ParsePatternSpecFile> can
be made available using

    use SlidingWindowSearch qw(ListMatches CountMatches ParsePatternSpecFile)
		
or alternatively

    use SlidingWindowSearch ':all'

=head1 SUPPORT

No support is offered.

=head1 TODO

=over 4

=item * Error handling

Error states should probably by signalled back to the calling program 
for handling, rather than the module generating fatal errors.

=item * Pattern width

Is there any simple way to get Perl to work out the width of a general 
(fixed-width) regex from the regex itself?   Haven't thought of one...

=item * Too much strcpy'ing?

Might be sensible to pass references to C<$str> around, rather than copying
the whole string itself.  But a downside is that then constant strings, 
eg as used in the unit tests, need special attention; and if one tries
to hide the reference-ness from the user using subroutine prototypes, it
then becomes rather confusing that they can't call the subroutine with
a constant string...

=item * POD formatting issues

Some formatting glitches being produced by pod2html.  Not sure why
they're happening or how to sort them out...

=back 

=head1 AUTHOR

James Heald E<lt>j.heald@ucl.ac.ukE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright 2012 by James Heald

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

