#!/usr/bin/perl
use strict;
use warnings;
use Test::More;
use Test::Exception;
use Test::Warn;

# ACKNOWLEDGEMENT:
# A somewhat subtle bug that was leading to problems with the function calls
# was solved by an eagle-eyed contributor at PerlMonks:
# http://perlmonks.org/index.pl?node_id=1003647

BEGIN { use_ok('SlidingWindowSearch', qw(ListMatches CountMatches)); }

#
{
	SlidingWindowSearch::SetDebug(1); 
	my $debug=SlidingWindowSearch::GetDebug();
	is ($debug,		1,		'Debug mode turned on');
}
	
#	
{
	SlidingWindowSearch::SetDebug(0); 
	my $debug=SlidingWindowSearch::GetDebug();
	is ($debug,		0,		'Debug mode turned off');
}

## tests using dice and match method
# test extended pattern 'A' with overhang = 2
is(join(" ", SlidingWindowSearch::ListMatches_UsingGrep('ABRACADABRA','A',1,3)),
	'ABR BRA RAC ACA CAD ADA DAB ABR BRA',
	'ListMatches_UsingGrep correctly ran pattern A, extended case');
 
# test extended pattern 'R' with overhang = 2
is(join(" ", SlidingWindowSearch::ListMatches_UsingGrep('ABRACADABRA','R',1,3)),
	'ABR BRA RAC ABR BRA',
	'ListMatches_UsingGrep correctly ran pattern R, extended case');

## tests using sliding window method
# test hard-coded simple pattern '.A.'
is(join(" ", SlidingWindowSearch::ListMatches_HardCoded_Simple('ABRACADABRA')),
	'RAC CAD DAB',
	'Hard-coded pattern, simple case');

# test hard-coded extended pattern 'A' with an overhang of 2	
is(join(" ", SlidingWindowSearch::ListMatches_HardCoded_Extended('ABRACADABRA')),
	'ABR BRA RAC ACA CAD ADA DAB ABR BRA',
	'Hard-coded pattern, extended case');
	
# test simple pattern '.A.'
is(join(" ", ListMatches('ABRACADABRA','.A.')),
	'RAC CAD DAB',
	'ListMatches correctly ran pattern, simple case');
	
# test extended pattern 'A' with overhang = 2
is(join(" ", ListMatches('ABRACADABRA','A',1,3)),
	'ABR BRA RAC ACA CAD ADA DAB ABR BRA',
	'ListMatches correctly ran pattern A, extended case');
 
# test ListMatches dies okay with broken pattern
dies_ok {my @list = ListMatches('ABRACADABRA','.([A].',1,3)}
	'ListMatches dies correctly given broken pattern';

# test extended pattern 'R' with overhang = 2
is(join(" ", ListMatches('ABRACADABRA','R',1,3)),
	'ABR BRA RAC ABR BRA',
	'ListMatches correctly ran pattern R, extended case');

# test CountMatches with single pattern
{
	my %matchCount;
	my @patternList = qw(A);
	%matchCount = CountMatches('ABRACADABRA',@patternList,1,3);
	is_deeply (\%matchCount,
		{ABR => 1, ACA => 1, ADA => 1, BRA => 1, CAD => 1, DAB => 1, RAC => 1},
		'CountMatches correctly ran pattern A, extended case');
}

# test CountMatches with two patterns
{
	my %matchCount;
	my @patternList = qw(A R);
	%matchCount = CountMatches('ABRACADABRA',@patternList,1,3);
	is_deeply (\%matchCount,
		{ABR => 2, ACA => 1, ADA => 1, BRA => 2, CAD => 1, DAB => 1, RAC => 2},
		'CountMatches correctly ran patterns "A" + "R", extended case');
}

# test CountMatches with two patterns, plus updating an existing count
{
	my %matchCount;
	$matchCount{'TEST'} = 1;
	my @patternList = qw(A R);
	%matchCount = CountMatches('ABRACADABRA',@patternList,1,3,undef,%matchCount);
	is_deeply (\%matchCount,
		{ABR => 2, ACA => 1, ADA => 1, BRA => 2, CAD => 1, DAB => 1, RAC => 2, TEST => 1},
		'CountMatches correctly updated an existing count');
}

# test CountMatches dies if the sequence has a space in it
{
	my %matchCount;
	my @patternList = qw(A);
	dies_ok {%matchCount = CountMatches('ABRA ADABRA',@patternList,1,3)}
		'Dies if the sequence has a space in it';
}

# test over-ruling CheckSeq
SlidingWindowSearch::SetChecks(0);
{
	my %matchCount;
	my @patternList = qw(A);
	%matchCount = CountMatches('ABRA ADABRA',@patternList,1,3);
	is_deeply (\%matchCount,
		{ABR => 1, 'A A' => 1, ADA => 1, BRA => 1, ' AD' => 1, DAB => 1, 'RA ' => 1},
		'CountMatches runs correctly if CheckSeq over-ruled');
}


# test CountMatches dies if the sequence has a \r in it
SlidingWindowSearch::SetChecks(1);
{
	my %matchCount;
	my @patternList = qw(A);
	dies_ok {%matchCount = CountMatches("ABRA\rADABRA",@patternList,1,3)}
		'Dies okay again if the sequence has a \r in it';
}
	
# test CountMatches with 'UsingGrep' algorithm
### Note: doesn't actually check 'UsingGrep' was used, just that *something* returned the right answer
### (though if it wasn't 'UsingGrep' then there should be a warning)
{
	my %matchCount;
	my @patternList = qw(A R);
	%matchCount = CountMatches('ABRACADABRA',@patternList,1,3,'UsingGrep');
	is_deeply (\%matchCount,
		{ABR => 2, ACA => 1, ADA => 1, BRA => 2, CAD => 1, DAB => 1, RAC => 2},
		'CountMatches runs patterns "A" + "R" correctly, with UsingGrep');
}

# test CountMatches reverts to default okay if a nonsense algorithm is specified
{
	my %matchCount;
	my @patternList = qw(A R);
	(warning_is {%matchCount = CountMatches('ABRACADABRA',@patternList,1,3,'FishWife')}
		{carped => "Method 'FishWife' not recognised; reverting to default"},
		"CountMatches warns if method is non-existent") &&
	is_deeply (\%matchCount,
		{ABR => 2, ACA => 1, ADA => 1, BRA => 2, CAD => 1, DAB => 1, RAC => 2},
		'... and correctly reverts to default');
}
	
done_testing();

